# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory
import os

from temporalio import activity
from temporalio.activity import logger

from jax.cs.storage import StorageObject
from jax.pyjit.dao import StorageKey


from pathlib import Path
import shutil

###############################################################################
# These activities are provided by universal activities when deployed.
# We have test versions here to make sure they are implemented in test mode.
# We use the term 'dummy' to indicate these are not the real activities.
###############################################################################


@activity.defn(name="io-temporal-download-object")
async def dummy_download_object(
    key: StorageKey, run_id: str, overwrite: bool, ignoreNotExisting: bool
) -> str:

    try:
        dir: Path = get_dir(run_id)
        file: Path = Path(dir, key.object)
        os.makedirs(file.parent, exist_ok=True)
        file.touch()

        uri: str = key.to_uri()
        wrapped = StorageObject(uri)

        with open(file, "wb") as download_file:
            data = wrapped.download_as_bytes()
            download_file.write(data)

        return str(file)
    except Exception as e:
        logger.error(e)
        raise e


@activity.defn(name="io-temporal-upload")
async def dummy_upload(key: StorageKey, file: str, replace: bool) -> str:

    try:
        if not os.path.exists(file):
            return

        local_wrapped: StorageObject = StorageObject(file)
        uri: str = key.to_uri(
            create=True
        )  # Test uri will create local file, otherwise, create does nothing.
        remote_wrapped: StorageObject = StorageObject(uri)
        remote_wrapped.replace(local_wrapped)
        return uri

    except Exception as e:
        logger.error(e)
        raise e


@activity.defn(name="io-temporal-clean")
async def dummy_clean(run_id: str) -> None:

    try:
        dir: Path = get_dir(run_id)
        shutil.rmtree(dir)

    except Exception as e:
        logger.error(e)
        raise e


def get_dir(run_id: str) -> Path:

    CACHE_DIR: str = os.getenv("CACHE_DIR", "cache")
    cache: Path = Path(CACHE_DIR)
    dir: Path = Path(cache, run_id)
    # Make dirs if not existing
    os.makedirs(dir, exist_ok=True)
    return dir



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory
try:
    import pkg_resources

    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    __path__ = pkgutil.extend_path(__path__, __name__)

import os

from typing import List

class SharedConfig:
    """Config and Settings Configuration."""

    DEBUG: bool = bool(os.getenv("DEBUG", None))
    TESTING: bool = bool(os.getenv("TESTING", None))

    USE_NIO_STORAGE: bool = bool(os.getenv("USE_NIO_STORAGE", None))
    NIO_STORAGE_ROOT: str = os.getenv("NIO_STORAGE_ROOT", None)

    LOG_LEVEL: str = os.getenv("LOG_LEVEL", "WARNING")
    TEMPORAL_NAMESPACE: str = os.getenv("TEMPORAL_NAMESPACE", "default")
    TEMPORAL_URI: str = os.getenv(
        "TEMPORAL_URI", "localhost:7233"
    )  # Deploy should set to CS cluster location.
    CACHE_DIR: str = os.getenv("CACHE_DIR", "cache")

    TEST_BUCKET_NAME: str = os.getenv("TEST_BUCKET_NAME", None)

    DOWNLOAD_TIMEOUT: int = int(os.getenv("DOWNLOAD_TIMEOUT", "1200"))
    DOWNLOAD_ATTEMPTS: int = int(os.getenv("DOWNLOAD_ATTEMPTS", "4"))

    UPLOAD_TIMEOUT: int = int(os.getenv("UPLOAD_TIMEOUT", "1200"))
    UPLOAD_ATTEMPTS: int = int(os.getenv("UPLOAD_ATTEMPTS", "4"))

    CLEAN_TIMEOUT: int = int(os.getenv("CLEAN_TIMEOUT", "300"))
    CLEAN_ATTEMPTS: int = int(os.getenv("CLEAN_ATTEMPTS", "4"))

    # IO operations which come from other workers.
    IO_STORE_QUEUE = os.getenv("STORE_QUEUE", "STORE_TASK_QUEUE")
    IO_DOWNLOAD_OBJECT = os.getenv("IO_DOWNLOAD_OBJECT", "io-temporal-download-object")
    IO_UPLOAD = os.getenv("IO_UPLOAD", "io-temporal-upload")
    IO_CLEAN = os.getenv("IO_CLEAN", "io-temporal-clean")


shared_config = SharedConfig()


def set_test_defaults():
    import os

    os.environ["STORE_QUEUE"] = "TEST"
    os.environ["DOWNLOAD_TIMEOUT"] = "30"
    os.environ["DOWNLOAD_ATTEMPTS"] = "1"
    os.environ["UPLOAD_TIMEOUT"] = "30"
    os.environ["UPLOAD_ATTEMPTS"] = "1"
    os.environ["CLEAN_TIMEOUT"] = "30"
    os.environ["CLEAN_ATTEMPTS"] = "1"

    shared_config.STORE_QUEUE = "TEST"
    shared_config.DOWNLOAD_TIMEOUT = 30
    shared_config.DOWNLOAD_ATTEMPTS = 1
    shared_config.UPLOAD_TIMEOUT = 30
    shared_config.UPLOAD_ATTEMPTS = 1
    shared_config.CLEAN_TIMEOUT = 30
    shared_config.CLEAN_ATTEMPTS = 1


class Settings:
    """Settings class for Auth Client."""

    AUTH_DOMAIN: str = "thejacksonlaboratory.auth0.com"
    AUTH_CLIENT_ID: str = "f8QZPcIrPIG6DIeWR2Rr3C8X5bzx8zBz"
    AUTH_ALGORITHMS: List[str] = ["RS256"]
    AUTH_SCOPES: List[str] = ["openid", "profile", "email"]
    AUTH_AUDIENCE: str = "https://cube.jax.org"

auth_settings = Settings()


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory

from typing import Optional, Any, Dict
from .dao import UntypedStatus, Input, ToolConf, StorageKey
from jax.pyjit.auth import get_access_token

from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
import base64


import sys
import os
import time
import requests
import random
import string
from pathlib import Path
import json
from time import sleep

def run_workflow(input: Input, 
                 tool: ToolConf,
                 server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
                 query_interval=2,
                 subscriber=None) -> UntypedStatus:
    """
    Run a workflow given an input and tool configuration.
    """
    
    # Submit the workflow
    monitor_uri = submit_workflow(input, tool, server_uri, query_interval)
    
    # Monitor until complete
    return monitor_workflow(monitor_uri, query_interval)

def monitor_workflow(monitor_uri: str, query_interval=2, subscriber=None) -> UntypedStatus:
    """
    Monitor a workflow until completion.
    """
    
    status: UntypedStatus = None
    
    # Monitor workflow status
    while not is_final(status):
        try:
            token = get_access_token()
            headers = {
                "Authorization": f"Bearer {token}",
                "Content-Type": "application/json"
            }
            response = requests.get(monitor_uri, headers=headers)
            response.raise_for_status()
            status: UntypedStatus = response.json()
            
            if subscriber is not None:
                subscriber(status)
            
            complete: float = status["complete"]
            complete_percent: float = complete * 100.0
            state:str = status["state"]
            message: str = status["message"] 
            sys.stdout.write(f"\r{state} {complete_percent}% {message}")
            sys.stdout.flush()
            if is_final(status):
                break
            
            time.sleep(query_interval)
           
        except Exception:
            break

    return status


def submit_workflow(input: Input, 
                 tool: ToolConf,
                 server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
                 query_interval=2) -> Any:
    """
    Submit a workflow given an input and tool configuration.
    """
  
    instr = create_instruction(input)
    token = get_access_token()
    
    url = f"{server_uri}/process"
    params = {"type": tool.name}
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }
    with requests.post(url, params=params, json=instr, headers=headers) as response:
        response.raise_for_status()  # Raises an error for bad responses
        monitor: str = response.json()  # Or response.text if you expect plain text
        monitor_uri = monitor["response"]
        print(f"Monitor URI: {monitor_uri}")
        sleep(2)  # Give server a moment to start processing, not strictly necessary
                  # but gives a better user experience.
        return monitor_uri
    
def delete(file_or_dir: StorageKey,
           server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
           type: str = "File") -> None:
    """
    Delete a file or directory.
    """
  
    blob: Dict = create_blob(file_or_dir, type)
    token: str = get_access_token()
    
    url = f"{server_uri}/delete"
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }
    
    with requests.post(url, json=blob, headers=headers) as response:
        response.raise_for_status()  # Raises an error for bad responses


def download(file: StorageKey,
           server_uri: Optional[str] = "https://imagetools-dev.jax.org/api",
           dir: Path = Path("./tmp")) -> Path:
    """
    Delete a file or directory.
    """
  
    blob: Dict = create_blob(file)
    token: str = get_access_token()
    
    url = f"{server_uri}/download"
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json"
    }
    os.makedirs(dir, exist_ok=True)
    filename = os.path.basename(file.object)
    output_path = os.path.join(dir, filename)
    
    params = {"info": json.dumps(blob)}
    with requests.get(url, params=params, headers=headers, stream=True) as response:
        response.raise_for_status()
        with open(output_path, "wb") as f:
            for chunk in response.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
    return Path(output_path)

def upload(file: Path,
           dir: StorageKey,
           server_uri: Optional[str] = "https://imagetools-dev.jax.org/api") -> Path:
    """
    Delete a file or directory.
    """
  
    token: str = get_access_token()
    
    url = f"{server_uri}/upload"
    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
        "uri" : dir.to_uri(),
        "key" : get_upload_key(file)
    }
    
    with open(file, "rb") as f:
        files = {"file": (file.name, f)}
        with requests.post(url, headers=headers, files=files) as response:
            response.raise_for_status()
        
    return file


UPLOAD_KEY = os.getenv("UPLOAD_KEY", "A secret key base64 encoded==")

def get_upload_key(file_path: Path) -> str:
    key = get_key()

    # AES in ECB mode (since Java Cipher.getInstance("AES") defaults to AES/ECB/PKCS5Padding)
    cipher = Cipher(algorithms.AES(key), modes.ECB(), backend=default_backend())
    encryptor = cipher.encryptor()

    # Prepare plaintext (filename as UTF-8)
    plaintext = file_path.name.encode("utf-8")

    # Apply PKCS7 padding (block size = 128 bits for AES)
    padder = padding.PKCS7(128).padder()
    padded = padder.update(plaintext) + padder.finalize()

    # Encrypt
    encrypted_bytes = encryptor.update(padded) + encryptor.finalize()

    # Return base64 string
    return base64.b64encode(encrypted_bytes).decode("utf-8")
   
def get_key():
    # Decode the base64 key into raw bytes
    key_bytes = base64.b64decode(UPLOAD_KEY)
    return key_bytes

def create_instruction(input: Input) -> Dict:
    instr = {
        "file": create_blob(input.image),
        "request": input.request,
    }
    return instr

def create_blob(storage_key, type: str = "File") -> Dict:
    blob = {
        "name": os.path.basename(storage_key.object),
        "project": {
            "bucket": storage_key.bucket,
            "name": "Python api submission to {}".format(storage_key.bucket),
        },
        "type": type,
        "relPath": storage_key.object,
    }
    return blob
    

def is_final(status: UntypedStatus) -> bool:
    if status is None:
        return False
    state:str = status["state"]
    return state in ("COMPLETE", "ERROR", "CANCELLED")

def is_storage_result(result: str) -> bool:
    
    if "://" in result:
        return True
    return False

def random_string(size: Optional[int] = 4) -> str:
    # Generate a n-character random string (uppercase + lowercase letters)
    rand_str = ''.join(random.choice(string.ascii_letters) for _ in range(size))
    return rand_str



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Copyright [2025] The Jackson Laboratory


from dataclasses import dataclass, field
from typing import List, Optional, Any
from pathlib import Path
from enum import Enum
from urllib.parse import urlparse

import os

from jax.pyjit import shared_config



class State:
    SUBMITTED: str = "SUBMITTED"
    RUNNING: str = "RUNNING"
    COMPLETE: str = "COMPLETE"
    ERROR: str = "ERROR"
    NONE: str = "NONE"
    CANCELLED: str = "CANCELLED"


@dataclass
class UntypedStatus:
    """
    Based on cimg-api/org.jax.cimg.api.evt.UntypedStatus
    which we use in Image Tools.
    These fields are shared across multiple tools.
    Do not change them.
    """

    message: Optional[str] = None
    state: str = State.NONE
    complete: float = 0
    errorStack: Optional[str] = None
    workflowId: Optional[str] = None
    results: List[str] = field(default_factory=lambda: [""])
    submissionInput: Optional[str] = None
    userName: Optional[str] = None


class Protocol(str, Enum):
    GS = "GS"
    S3 = "S3"
    NIO = "NIO"


@dataclass
class StorageKey:

    bucket: str
    object: str
    protocol: Optional[str] = None # One of Protocol
    endpoint: Optional[str] = None
    test: Optional[bool] = False
    
    def to_dict(self):
        # This is require to make 
        # json serialization to work.
        return {
            "bucket": self.bucket,
            "object": self.object,
            "protocol": self.protocol,
            "endpoint": self.endpoint,
        }
        
    def clone(self) -> "StorageKey":
        # You can trick python here using a string to let
        # it return a type but you cannot return a StorageKey
        # because we have not defined it yet.
        # Now wash your eyes.
        return StorageKey(
            bucket=self.bucket,
            object=self.object,
            protocol=self.protocol,
            endpoint=self.endpoint,
            test=self.test,
        )

    def set_file_name(self, file_name: str) -> "StorageKey":
        """
        Replace the file name in the object path with the given file_name.
        """
        p = Path(self.object)
        if p.parent == Path('.'):
            self.object = file_name
        else:
            if p.is_dir() or self.object.endswith('/'):
                self.object = str(p / file_name)
            else:
                self.object = str(p.parent / file_name)
            
        return self
            
    def to_uri(self, create=False) -> str:
        if self._is_local():
            file_path: str = None
            if self.endpoint is None:
                file_path: str = "{}/{}".format(self.bucket, self.object)
            else:
                file_path: str = "{}/{}/{}".format(
                    self.endpoint, self.bucket, self.object
                )

            NIO_STORAGE_ROOT: str = shared_config.NIO_STORAGE_ROOT
            if NIO_STORAGE_ROOT:
                file_path = "{}/{}".format(NIO_STORAGE_ROOT, file_path)

            # Remove double slashes
            file_path = file_path.replace("//", "/")

            if create:
                path: Path = Path(file_path)
                os.makedirs(path.parent, exist_ok=True)
                path.touch(exist_ok=True)
            return file_path
        else:
            return f"{self.protocol.lower()}://{self.bucket}/{self.object}"

    def _is_local(self) -> bool:
        if self.protocol == Protocol.NIO:
            return True
        if self.test:
            return True

        USE_NIO_STORAGE: bool = shared_config.USE_NIO_STORAGE
        if USE_NIO_STORAGE:
            return True
        return False

    @classmethod
    def from_uri(cls, uri: str) -> "StorageKey":
        """
        Parse a URI and return a StorageKey instance.
        Supports formats like:
          - gs://bucket/object
          - s3://bucket/object
          - nio://bucket/object
          - /local/path/to/object (NIO/local)
        """
        parsed = urlparse(uri)
        if parsed.scheme in ("gs", "s3", "nio"):
            protocol = parsed.scheme.upper()
            bucket = parsed.netloc
            object_ = parsed.path.lstrip("/")
            return cls(bucket=bucket, object=object_, protocol=protocol)
        else:
            # Assume local path (NIO)
            # You may want to split the path differently depending on your conventions
            parts = Path(uri).parts
            if len(parts) >= 2:
                bucket = parts[0]
                object_ = str(Path(*parts[1:]))
            else:
                bucket = ""
                object_ = uri
            return cls(bucket=bucket, object=object_, protocol=Protocol.NIO)

# @see org.jax.cimg.api.workflow.AbstactRequest
# Your request should inherit from this base class please.
class AbstractRequest:
    info: Optional[Any] = None

@dataclass
class Input:
    """
    Input is the single class we send to any workflow in JIT.
    request is the actual request object specific to the workflow.
    """
    image: StorageKey
    request: AbstractRequest
    result: Optional[Any] = None
    userName: Optional[str] = None
    # Tried using field(default=None, metadata=config(field_name="userName")) to make is user_name 
    # but did not work.

class ToolConf:
    """
    Class containing queue name constants.
    """    
    def __init__(self, 
                 name: Optional[str], 
                 queue: Optional[str], 
                 workflow_name: Optional[str], 
                 workflow_query: Optional[str], 
                 namespace: Optional[str] = "cimg-dev"):
        
        self.name = name
        self.queue = queue
        self.workflow_name = workflow_name
        self.workflow_query = workflow_query
        self.namespace = namespace
        
class Tools:
    
    # Not a ProcessType value
    NONE: ToolConf = ToolConf("NONE", "NONE", None, None)
    
    #-------------------------
    # NOTE These strings must match org.jax.cimg.api.dao.ProcessType
    # precisely.
    #-------------------------
    CROP: ToolConf = ToolConf("CROP", "SEGMENTATION_TASK_QUEUE", "Segmentation Workflow", "Segmentation Query")
    FIND: ToolConf = ToolConf("FIND", "FIND_TASK_QUEUE", "Find Workflow", "Find Query")
    HTQUANT: ToolConf = ToolConf("HTQUANT", "HTQUANT_QUEUE", "HtQuant Workflow", "HtQuant Query")
    ALIGNMENT: ToolConf = ToolConf("ALIGNMENT", "SIFT_QUEUE", "Sift Workflow", "Sift Query")
    EDOF: ToolConf = ToolConf("EDOF", "EDOF_QUEUE", "EDOF Workflow", "EDOF Query")
    KIDNEY_CLASSIFIER: ToolConf = ToolConf("KIDNEY_CLASSIFIER", "orchestration-task", "OrchestrationWorkflow", "classifierStatus")
    DECONVOLUTION: ToolConf = ToolConf("DECONVOLUTION", "DECONVOLUTION_QUEUE", "Color Deconvolution Workflow", "Color Deconvolution Query")
    YOLO_SEGDETECT: ToolConf = ToolConf("YOLO_SEGDETECT", "YOLO_SEGDETECT_QUEUE", "YoloSegdetect Workflow", "YoloSegdetect Query")
    RETINAL_LAYER: ToolConf = ToolConf("RETINAL_LAYER", "RETINAL_LAYER_QUEUE", "Retinal Layer Workflow", "Retinal Layer Query")
    STAIN_ADAPTER: ToolConf = ToolConf("STAIN_ADAPTER", "STAIN_ADAPTER_QUEUE", "Stain Adapter Workflow", "Stain Adapter Query")
    CONVERT: ToolConf = ToolConf("CONVERT", "CONVERT_TASK_QUEUE", "Convert Workflow", "Convert Query") 


def blank_key() -> StorageKey:
    return StorageKey("", "", test=True)


@dataclass
class AbstractRequest:
    info: str

    def __eq__(self, other):
        if not isinstance(other, AbstractRequest):
            return False
        return self.info == other.info

    def __hash__(self):
        return hash(self.info)


@dataclass
class DeconvolutionRequest(AbstractRequest):
    inputPath: Optional[StorageKey] = None

    stain1: List[float] = field(default_factory=lambda: [0.563, 0.720, 0.406])
    stain2: List[float] = field(default_factory=lambda: [0.216, 0.801, 0.558])

    stain1Max: float = 2.0
    stain2Max: float = 1.0

    alpha: int = 1
    beta: float = 0.15
    intensityNorm: int = 240
    grayscale: bool = False

    stain1ImageOutput: Optional[StorageKey] = None
    stain2ImageOutput: Optional[StorageKey] = None
    normalizedImageOutput: Optional[StorageKey] = None
    imageStackOutput: Optional[StorageKey] = None

    def __post_init__(self):
        # Ensure info is set to DECONVOLUTION by default
        if not self.info:
            self.info = Tools.DECONVOLUTION.name



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
from typing import Dict
from .dao import Tools, ToolConf


def _create_template_requests() -> Dict:
    
    # TODO Do we want to have proper concrete types for these requests?
    # We could autogenerate them from the Java classes as we did for the
    # typescript JIT client.
    
    # We could also move the existing dao concrete types in all the python
    # tools to dao here then import them into each tool.
    # TODO Submit a ticket to do that.
    
    ret = {}
    
    # Crop
    request = {
        "info": "CROP",
        
        # This is the StorageKey of the folder to be
        # used to hold the results.
        "outputDir": None, # StorageKey
        "bytesPerPixel": 0,
        "cropIndices": None,
        
        # Use this to customize the size of
        # things being detected.
        "particleSettings": {
            "min": 1200000,
            "max": 2147483647,
            "cMin": 0.1,
            "cMax": 1,
            "hardBounds": True,
        },
        
        # This is used to process the image
        # prior to segementation i.e. making the mask.
        "imagePreprocess": {
            "rollingBallRadius": 50,
            "sigma": 100,
            "accuracy": 0.002,
            "autoThreshold": "Otsu",
        },
        "findResolution": -2147483648,
        "optimumFindSize": 16000000,
        "cropResolution": 0,
        
        # The settings for regions
        "regionSettings": {
            "minimumAspectRatio": 0.15,
            "margin": 10,
            "findRegions": True,
            "showRegions": False,
            "reference": None,
            "manualRegionResolution": 0,
            "manualRegions": [],
            "userRegions": None,
            "generateProofImage": True,
            "generatePreviews": True,
            "previewSize": 2048,
            "zip": True,
            "saveAsJson": True,
            "fontSize": 160,
            "cropFoundRegionsAsRectangles": True,
            "forceSquare": False,
            "cropAsGreyScale": False,
            "order": "ROWS",
        },
        "prefix": "",
        
        # Tiling is being used setup here.
        "tileConfiguration": {
            "tileSize": [4096, 4096],
            "order": "IMAGE",
            "clip": True,
            "alwaysTile": False,
            "expandCroppedFiles": True,
            "cropFillColor": None,
        },
        
        # Save the request generated, useful for 
        # using this run with the API later.
        "saveRequest": True,
        "imageJPlugin": False,
    }
    ret[Tools.CROP] = request
    
    # TODO Use concrete python class?
    request = {
        "info" : "DECONVOLUTION",
        "inputPath" : None,
        "stain1" : [ 0.563, 0.72, 0.406 ],
        "stain2" : [ 0.216, 0.801, 0.558 ],
        "stain1Max" : 2.0,
        "stain2Max" : 1.0,
        "alpha" : 1,
        "beta" : 0.15,
        "intensityNorm" : 240,
        "grayscale" : False,
        "stain1ImageOutput" : None,
        "stain2ImageOutput" : None,
        "normalizedImageOutput" : None,
        "imageStackOutput" : None
    }
    ret[Tools.DECONVOLUTION] = request

    # TODO Use concrete python class?
    request = {
      "info" : "YOLO_SEGDETECT",
      "inputPaths" : None,
      "weightsPath" : None,
      "outputPath" : None,
      "downsamplingFactor" : 5,
      "visualize" : False,
      "overlapX" : 0,
      "overlapY" : 0,
      "confidence" : 0.6,
      "iouThreshold" : 0.5,
      "nmsThreshold" : 0.3,
      "saveSegment" : False,
      "level" : 2,
      "db_upsert" : False
    }
    ret[Tools.YOLO_SEGDETECT] = request

    # TODO Use concrete python class?
    request = {
      "info" : "ALIGNMENT",
      
      # StorageKeys
      "align" : None,
      "reference" : None,
      "shg" : None,
      "siftOutput" : None,
      "resizedN" : None,
      "resizedSF" : None,
      "overlayedResult" : None,
      
      # Config
      "transform" : True,
      "flip" : True,
      "downsample" : False,
      "gamma" : 1.0,
    }
    ret[Tools.ALIGNMENT] = request
    
    request = {
        "info" : "EDOF",
      
        # Parameters:
        # -gradient_kernel=(5)
        #        Size of kernel used for computation of image gradients.  Must be 1, 3, 5, 7.
        "gradientKernel" : 5,
        
        # -noise_filter=(3)
        #    Size of median filter used to reduce noise in the z-map. Must be 0, 3 or 5
        "imageNoiseFilter" : 3,

        # -noise_filter=(3)
        #      Size of median filter used to reduce noise in the z-map. Must be 0, 3 or 5
        "zmapNoiseFilter" : 3,

        # -low_pass=(2)
        #    zmap lowpass filter object size.
        "lowPass" : 2,
        
        # It will be hard to know this from the client side
        # because the analysis may run anywhere in temporal.io 
        # distributed computing environment.
        "threads" : 2,
        "dryRun" : False,
        
        # A list of StorageKeys for the input images.
        # It is a list because edof can run either on one
        # image which contains a stack or muliple images 
        # in a directory.
        "input" : [],
        
        # -image=("edof.tif") 
        # StorageKey for the output image file.  Most standard formats are supported.
        "outputImage" : None,
        
        # -zmap=("zmap.tif")
        # StorageKey for the z-map file saved as a 16-bit monochrome image.
        "outputZmap" : None,
      
        # -zmap_input=(none)
        #    StorageKey which when specified, this z-map is used for extraction of the final image.
        "inputZmap" : None,
    }
    ret[Tools.EDOF] = request

    
    return ret

template_request_dict: Dict = _create_template_requests()

def create_template_request(type: ToolConf):
    if type not in template_request_dict:
        raise ValueError(f"No template request for type {type.name}")
    return template_request_dict[type]
